#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import time
import random
import numpy as np
from pathlib import Path
from gurobipy import *

import pickle

import GridMap as gmap

PURSUER = 1
EVADER = 2

GREEDY = "GREEDY"
MONTE_CARLO = "MONTE_CARLO"
VALUE_ITERATION = "VALUE_ITERATION"

def compute_nash(matrix, only_value=False, minimize=False):
    maximizer_actions = matrix.shape[0]
    return 0, [1./maximizer_actions]*maximizer_actions

class Player:
    def __init__(self, robots, role, policy=GREEDY, color='r', epsilon=1, 
                 timeout=5.0, game_name=None):
        """ constructor of the Player class
        Args: robots: list((in,int)) - coordinates of individual player's robots
              role: PURSUER/EVADER - player's role in the game
              policy: GREEDY/MONTE_CARLO/VALUE_ITERATION - player's policy, 
              color: string - player color for visualization
              epsilon: float - [0,1] epsilon value for greedy policy
              timeout: float - timout for MCTS policy
              game_name: string - name of the currently played game 
        """
        # list of the player's robots
        self.robots = robots[:]
        # next position of the player's robots
        self.next_robots = robots[:]

        if role == "EVADER":
            self.role = EVADER
        elif role == "PURSUER":
            self.role = PURSUER
        else:
            raise ValueError('Unknown player role')

        # selection of the policy
        if policy == GREEDY:
            self.policy = self.greedy_policy
        elif policy == MONTE_CARLO:
            self.policy = self.monte_carlo_policy
            self.timeout = timeout * len(self.robots) # MCTS planning timeout
            self.tree = {}
            self.max_depth = 10
            self.step = 0
            self.max_steps = 100
            self.beta = 0.95
            self.c = 1
        elif policy == VALUE_ITERATION:
            self.policy = self.value_iteration_policy
            # values for the value iteration policy
            self.loaded_policy = None
            self.gamma = 0.95
        else:
            raise ValueError('Unknown policy')

        #parameters
        self.color = color # color for plotting purposes
        self.game_name = game_name # game name for loading vi policies


    #####################################################
    # Game interface functions
    #####################################################
    def add_robot(self, pos):
        """ method to add a robot to the player
        Args: pos: (int,int) - position of the robot
        """
        self.robots.append(pos)
        self.next_robots.append(pos)

    def del_robot(self, pos):
        """ method to remove the player's robot 
        Args: pos: (int,int) - position of the robot to be removed
        """
        idx = self.robots.index(pos)
        self.robots.pop(idx)
        self.next_robots.pop(idx)

    def calculate_step(self, gridmap, evaders, pursuers):
        """ method to calculate the player's next step using selected policy
        Args: gridmap: GridMap - map of the environment
              evaders: list((int,int)) - list of coordinates of evaders in the 
                            game (except the player's robots, if he is evader)
              pursuers: list((int,int)) - list of coordinates of pursuers in 
                       the game (except the player's robots, if he is pursuer)
        """        
        self.policy(gridmap, evaders, pursuers)

    def take_step(self):
        """ method to perform the step 
        """
        self.robots = self.next_robots[:]

    #####################################################
    #####################################################
    # GREEDY POLICY
    #####################################################
    #####################################################
    def greedy_policy(self, gridmap, evaders, pursuers, epsilon=1):
        """ Method to calculate the greedy policy action
        Args: gridmap: GridMap - map of the environment
              evaders: list((int,int)) - list of coordinates of evaders in the 
                            game (except the player's robots, if he is evader)
              pursuers: list((int,int)) - list of coordinates of pursuers in 
                       the game (except the player's robots, if he is pursuer)
              epsilon: float (optional) - optional epsilon-greedy parameter
        """
        self.next_robots = self.robots[:]

        # for each of player's robots plan their actions
        for idx in range(0, len(self.robots)):
            robot = self.robots[idx]

            # TODO: implement greedy policy

            ##################################################
            # RANDOM Policy
            ##################################################
            # extract possible coordinates to go (actions)
            neighbors = gridmap.neighbors4(robot)
            # introducing randomness in neighbor selection
            random.shuffle(neighbors)

            # select random goal
            pos_selected = neighbors[0]
            ##################################################

            # select the next action based on own role
            if self.role == PURSUER:
                self.next_robots[idx] = pos_selected

            if self.role == EVADER:
                self.next_robots[idx] = pos_selected

    #####################################################
    #####################################################
    # VALUE ITERATION POLICY
    #####################################################
    #####################################################
    def init_values(self, gridmap):
        mapping_i2c = {}
        mapping_c2i = {}
        count = 0
        for i in range(gridmap.width):
            for j in range(gridmap.height):
                if gridmap.passable((i, j)):
                    mapping_i2c[count] = (i, j)
                    mapping_c2i[(i, j)] = count
                    count += 1
        return mapping_i2c, mapping_c2i, count

    def random_policy(self, coord_state, gridmap, mapping_c2i, role):
        a, b, c = coord_state
        neigh_a = gridmap.neighbors4(a)
        neigh_b = gridmap.neighbors4(b)
        neigh_c = gridmap.neighbors4(c)        
        if role == PURSUER:
            combined_actions = []
            for action_one in neigh_b:
                for action_two in neigh_c:
                    combined_actions.append((mapping_c2i[action_one], mapping_c2i[action_two]))
            return(combined_actions, [1/len(combined_actions)]*len(combined_actions))
        else:
            combined_actions = []
            for action in neigh_a:
                combined_actions.append(mapping_c2i[action])
            return(combined_actions, [1/len(combined_actions)]*len(combined_actions))

    def compute_random_policy(self, gridmap):
        mapping_i2c, mapping_c2i, spaces = self.init_values(gridmap)
        values = np.zeros((spaces,spaces,spaces))
        evader_policy = {}
        pursuer_policy = {}
        for a in range(spaces):
            for b in range(spaces):
                for c in range(spaces):
                    coord_state = (mapping_i2c[a], mapping_i2c[b], mapping_i2c[c])
                    evader_policy[(a,b,c)] = self.random_policy(coord_state, gridmap, mapping_c2i, EVADER)
                    pursuer_policy[(a,b,c)] = self.random_policy(coord_state, gridmap, mapping_c2i, PURSUER)
        return values, evader_policy, pursuer_policy, mapping_i2c, mapping_c2i

    def value_iteration_policy(self, gridmap, evaders, pursuers):
        """
        Method to calculate the value-iteration policy action

        Parameters
        ----------
        gridmap: GridMap
            Map of the environment
        evaders: list((int,int))
            list of coordinates of evaders in the game (except the player's robots, if he is evader)
        pursuers: list((int,int))
            list of coordinates of pursuers in the game (except the player's robots, if he is pursuer)
        """
        self.next_robots = self.robots[:]

        # if there are not precalculated values for policy
        if not self.loaded_policy:
            policy_file = Path("policies/" + self.game_name + ".policy")
            ###################################################
            # if there is policy file, load it...
            ###################################################
            if policy_file.is_file():
                # load the strategy file
                self.loaded_policy = pickle.load(open(policy_file, 'rb'))
                ###################################################
            # ...else calculate the policy
            ###################################################
            else:
                values, evader_policy, pursuer_policy, mapping_i2c, mapping_c2i = self.compute_random_policy(gridmap)

                self.loaded_policy = (values, evader_policy, pursuer_policy, mapping_i2c, mapping_c2i)

                pickle.dump(self.loaded_policy, open(policy_file, 'wb'))

        values, evader_policy, pursuer_policy, mapping_i2c, mapping_c2i = self.loaded_policy

        if self.role == PURSUER:
            state = (mapping_c2i[evaders[0]], mapping_c2i[self.robots[0]], mapping_c2i[self.robots[1]])
        else:
            state = (mapping_c2i[self.robots[0]], mapping_c2i[pursuers[0]], mapping_c2i[pursuers[1]])

        if self.role == PURSUER:
            action_index = np.random.choice(tuple(range(len(pursuer_policy[state][0]))), p=pursuer_policy[state][1])
            action = pursuer_policy[state][0][action_index]
            self.next_robots[0] = mapping_i2c[action[0]]
            self.next_robots[1] = mapping_i2c[action[1]]
        else:
            action_index = np.random.choice(tuple(range(len(evader_policy[state][0]))), p=evader_policy[state][1])
            action = evader_policy[state][0][action_index]
            self.next_robots[0] = mapping_i2c[action]
            #####################################################
